/*
 * Copyright (c) 1997 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 */

#include "life.h"


static char *	SafeAlloc PROTO((int));


/*
 * Allocate a new object structure.
 * It contains one row element, the termrow.
 * We quit if the allocation fails.
 */
OBJECT *
AllocateObject()
{
	OBJECT *obj;

	obj = freeobjects;

	if (obj)
		freeobjects = obj->next;
	else
	{
		obj = newobjects;

		if (obj >= endobjects)
		{
			obj = (OBJECT *) SafeAlloc(sizeof(OBJECT) * ALLOCOBJ);

			newobjects = obj;
			endobjects = obj + ALLOCOBJ;
		}

		newobjects++;
	}

	obj->next = NULL;
	obj->firstrow = termrow;
	obj->lastrow = NULL;
	obj->usedrows = termrow;
	obj->reserved = reserve;
	obj->name[0] = '\0';
	obj->count = 0;
	obj->gen = 0;
	obj->currow = 0;
	obj->curcol = 0;
	obj->origrow = 0;
	obj->origcol = 0;
	obj->cellmarks = 0;
	obj->update = 0;
	obj->viewrows = 0;
	obj->viewcols = 0;
	obj->autoscale = FALSE;
	obj->tracks = defaulttracks;
	obj->frequency = defaultfrequency;
	obj->curmarks = MARK_GROUPS;

	memset(obj->wherelocs, 0, sizeof(obj->wherelocs));

	InitializePath(obj);
	SetScale(obj, defaultscale);

	return obj;
}


/*
 * Allocate a new row structure.
 * It contains one row element, the termcell.
 * We quit if the allocation fails.
 */
ROW *
AllocateRow()
{
	ROW *	rp;

	rp = freerows;

	if (rp)
		freerows = rp->next;
	else
	{
		rp = newrows;

		if (rp >= endrows)
		{
			rp = (ROW *) SafeAlloc(sizeof(ROW) * ALLOCROW);

			newrows = rp;
			endrows = rp + ALLOCROW;
		}

		newrows++;
	}

	rp->next = NULL;
	rp->firstcell = termcell;
	rp->lastcell = NULL;
	rp->count = 0;

	return rp;
}


/*
 * Allocate a new cell structure.
 * We quit if the allocation fails.
 */
CELL *
AllocateCell()
{
	CELL *	cp;

	cp = freecells;

	if (cp)
	{
		freecells = cp->next;
		cp->next = NULL;
		cp->marks = MARK_ANY;

		return cp;
	}

	cp = newcells;

	if (cp >= endcells)
	{
		cp = (CELL *) SafeAlloc(sizeof(CELL) * ALLOCCELL);

		newcells = cp;
		endcells = cp + ALLOCCELL;
	}

	newcells++;
	cp->next = NULL;
	cp->marks = MARK_ANY;

	return cp;
}


/*
 * Free a list of rows along with all their cells.
 * The list is terminated either with a NULL or with the termrow element.
 * This returns the total number of cells that were freed.
 */
COUNT
FreeRows(firstrow)
	ROW *	firstrow;
{
	ROW *	rp;
	ROW *	lastrow;
	COUNT	count;

	lastrow = NULL;
	count = 0;

	/*
	 * Loop over all the rows, freeing the cells within them.
	 */
	for (rp = firstrow; rp && (rp != termrow); rp = rp->next)
	{
		if (rp->firstcell != termcell)
		{
			rp->lastcell->next = freecells;
			freecells = rp->firstcell;

			count += rp->count;
		}

		lastrow = rp;
	}

	/*
	 * Add the list of rows to the free row list.
	 */
	if (lastrow)
	{
		lastrow->next = freerows;
		freerows = firstrow;
	}

	return count;
}


/*
 * Copy a list of rows and cells.
 * Both the new and old lists are terminated with termrow entries.
 */
ROW *
CopyRows(ROW *oldrp)
{
	ROW *	rp;
	ROW *	retrp;
	ROW *	prevrp;
	CELL *	oldcp;
	CELL *	cp;
	CELL *	prevcp;

	retrp = NULL;
	prevrp = NULL;

	/*
	 * Loop over each row in the old list copying them to the new list.
	 */
	for (; oldrp != termrow; oldrp = oldrp->next)
	{
		rp = AllocateRow();
		rp->row = oldrp->row;

		if (prevrp)
			prevrp->next = rp;
		else
			retrp = rp;

		prevrp = rp;

		/*
		 * Copy all of the cells in the row.
		 */
		prevcp = NULL;

		for (oldcp = oldrp->firstcell;
			oldcp != termcell; oldcp = oldcp->next)
		{
			cp = AllocateCell();
			cp->col = oldcp->col;

			if (prevcp)
				prevcp->next = cp;
			else
				rp->firstcell = cp;

			prevcp = cp;
		}

		if (prevcp)
		{
			prevcp->next = termcell;
			rp->lastcell = prevcp;
		}
		else
			rp->firstcell = termcell;
	}

	if (prevrp == NULL)
		return termrow;

	prevrp->next = termrow;

	return retrp;
}


/*
 * Local malloc routine which exits if there is no memory.
 */
static char *
SafeAlloc(len)
	int	len;
{
	char *	cp;

	cp = malloc(len);

	if (cp)
		return cp;

	(*dev->close)(dev);

	fprintf(stderr, "Memory allocation failed\n");

	exit(1);
}

/* END CODE */
