/*
 * Copyright (c) 1997 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 */

#include <stdlib.h>
#include "life.h"


/*
 * List of line command routines.
 */
static	void	CmdDestroy PROTO((char *, VALUE, VALUE, BOOL, BOOL));
static	void	CmdEdit PROTO((char *, VALUE, VALUE, BOOL, BOOL));
static	void	CmdDumpInitData PROTO((char *, VALUE, VALUE, BOOL, BOOL));
static	void	CmdQuit PROTO((char *, VALUE, VALUE, BOOL, BOOL));
static	void	CmdAdvance PROTO((char *, VALUE, VALUE, BOOL, BOOL));
static	void	CmdFrequency PROTO((char *, VALUE, VALUE, BOOL, BOOL));
static	void	CmdObjects PROTO((char *, VALUE, VALUE, BOOL, BOOL));
static	void	CmdZero PROTO((char *, VALUE, VALUE, BOOL, BOOL));
static	void	CmdRead PROTO((char *, VALUE, VALUE, BOOL, BOOL));
static	void	CmdWrite PROTO((char *, VALUE, VALUE, BOOL, BOOL));
static	void	CmdTtyInput PROTO((char *, VALUE, VALUE, BOOL, BOOL));
static	void	CmdGrid PROTO((char *, VALUE, VALUE, BOOL, BOOL));
static	void	CmdNoGrid PROTO((char *, VALUE, VALUE, BOOL, BOOL));
static	void	CmdShape PROTO((char *, VALUE, VALUE, BOOL, BOOL));
static	void	CmdInsert PROTO((char *, VALUE, VALUE, BOOL, BOOL));
static	void	CmdClearPath PROTO((char *, VALUE, VALUE, BOOL, BOOL));
static	void	CmdRestorePath PROTO((char *, VALUE, VALUE, BOOL, BOOL));
static	void	CmdShortenPath PROTO((char *, VALUE, VALUE, BOOL, BOOL));
static	void	CmdCopy PROTO((char *, VALUE, VALUE, BOOL, BOOL));
static	void	CmdCopyMarked PROTO((char *, VALUE, VALUE, BOOL, BOOL));
static	void	CmdGet PROTO((char *, VALUE, VALUE, BOOL, BOOL));
static	void	CmdMove PROTO((char *, VALUE, VALUE, BOOL, BOOL));
static	void	CmdMoveMarked PROTO((char *, VALUE, VALUE, BOOL, BOOL));
static	void	CmdHelp PROTO((char *, VALUE, VALUE, BOOL, BOOL));
static	void	CmdRule PROTO((char *, VALUE, VALUE, BOOL, BOOL));
static	void	CmdEndInputLevel PROTO((char *, VALUE, VALUE, BOOL, BOOL));
static	void	CmdUndo PROTO((char *, VALUE, VALUE, BOOL, BOOL));
static	void	CmdRename PROTO((char *, VALUE, VALUE, BOOL, BOOL));
static	void	CmdSet PROTO((char *, VALUE, VALUE, BOOL, BOOL));
static	void	CmdVariables PROTO((char *, VALUE, VALUE, BOOL, BOOL));
static	void	CmdType PROTO((char *, VALUE, VALUE, BOOL, BOOL));
static	void	CmdUpdate PROTO((char *, VALUE, VALUE, BOOL, BOOL));
static	void	CmdWait PROTO((char *, VALUE, VALUE, BOOL, BOOL));
static	void	CmdMode PROTO((char *, VALUE, VALUE, BOOL, BOOL));
static	void	CmdUnMark PROTO((char *, VALUE, VALUE, BOOL, BOOL));
static	void	CmdInvertMarks PROTO((char *, VALUE, VALUE, BOOL, BOOL));
static	void	CmdAssign PROTO((char *, VALUE, VALUE, BOOL, BOOL));
static	void	CmdList PROTO((char *, VALUE, VALUE, BOOL, BOOL));
static	void	CmdDefault PROTO((char *, VALUE, VALUE, BOOL, BOOL));
static	void	CmdTracks PROTO((char *, VALUE, VALUE, BOOL, BOOL));
static	void	CmdPrefix PROTO((char *, VALUE, VALUE, BOOL, BOOL));
static	void	CmdGroups PROTO((char *, VALUE, VALUE, BOOL, BOOL));
static	void	CmdDelay PROTO((char *, VALUE, VALUE, BOOL, BOOL));


/*
 * Flags for line commands.
 */
#define	F_NONE		0x00	/* no special condition */
#define	F_NOARG		0x01	/* must not have an argument */
#define	F_ARG1		0x02	/* must have at least one argument */
#define	F_NORUN		0x04	/* illegal if running generations */
#define	F_STATUS	0x08	/* update status if command completes */
#define	F_UPDATE	0x10	/* update screen if command completes */
#define	F_REDRAW	0x20	/* redraw screen if command completes */
#define	F_ABBR		0x40	/* abbreviation works even if ambiguous */
#define	F_EXPAND	0x80	/* expand file names for this command */


/*
 * Dispatch table for line commands.  Those commands which are ambiguous
 * must be defined so as to be contiguous in the table.  A spaces delimits
 * the command itself from the help string for the command.
 */
typedef	struct
{
	char	*name;
	void	(*func) PROTO((char *, VALUE, VALUE, BOOL, BOOL));
	int	flags;
} CMDTAB;


static	CMDTAB	cmdtab[] =
{
	{"advance (marked cells) expr (gens)", CmdAdvance, F_NORUN|F_ABBR},
	{"assign (button) num (to macro) ch", CmdAssign, F_ARG1|F_NORUN},
	{"clearpath (from current object)", CmdClearPath, F_NOARG|F_NORUN},
	{"copy (current object to) obj", CmdCopy, F_ARG1|F_NORUN|F_ABBR},
	{"copymarked (cells to) obj", CmdCopyMarked, F_ARG1|F_NORUN},
	{"destroy (object named) obj", CmdDestroy, F_ARG1|F_NORUN},
	{"default param (to) expr", CmdDefault, F_ARG1|F_ABBR},
	{"delay (between updates) expr (msec)", CmdDelay, F_STATUS},
	{"dump (initialization data to) file", CmdDumpInitData, F_ARG1|F_NORUN},
	{"edit (object named) obj", CmdEdit, F_NORUN|F_UPDATE|F_ABBR},
	{"endinputlevel", CmdEndInputLevel, F_NOARG|F_STATUS},
	{"frequency (of typeout is) expr", CmdFrequency, F_STATUS},
	{"get (current object from) obj", CmdGet, F_ARG1|F_NORUN|F_ABBR},
	{"grid (character is) char", CmdGrid, F_REDRAW},
	{"groups (used for marks are) chars", CmdGroups, F_ARG1},
	{"help", CmdHelp, F_ABBR},
	{"insert (cells from) obj", CmdInsert, F_ARG1|F_NORUN|F_ABBR},
	{"invert (marks for all cells)", CmdInvertMarks, F_NOARG|F_NORUN|F_UPDATE},
	{"list (readable *.l files)", CmdList, F_UPDATE|F_ABBR},
	{"mode (for movement)", CmdMode, F_NORUN|F_STATUS},
	{"move (current object to) obj", CmdMove, F_ARG1|F_NORUN|F_UPDATE|F_ABBR},
	{"movemarked (cells to) obj", CmdMoveMarked, F_ARG1|F_NORUN|F_UPDATE},
	{"nogrid", CmdNoGrid, F_NOARG|F_REDRAW},
	{"objects (are listed)", CmdObjects, F_ABBR},
	{"prefix (for file paths is) path", CmdPrefix, F_NONE},
	{"quit (program)", CmdQuit, F_NOARG|F_ABBR},
	{"read (commands from) file", CmdRead, F_ARG1|F_EXPAND|F_NORUN|F_UPDATE|F_ABBR},
	{"rename (current object to) obj", CmdRename, F_ARG1|F_STATUS},
	{"restorepath (to current object)", CmdRestorePath, F_NOARG|F_NORUN},
	{"rule (for life are) born,live", CmdRule, F_NORUN|F_STATUS},
	{"set (variable) name (to) expr", CmdSet, F_ARG1|F_ABBR},
	{"shape (for drawing cells is) shape", CmdShape, F_REDRAW},
	{"shortenpath (from current object)", CmdShortenPath, F_NORUN},
	{"ttyinput", CmdTtyInput, F_UPDATE},
	{"tracks (for dead cells) action", CmdTracks, F_STATUS},
	{"type (value of expression) expr", CmdType, F_ARG1|F_STATUS|F_ABBR},
	{"undo (last change)", CmdUndo, F_NOARG|F_NORUN|F_UPDATE|F_ABBR},
	{"unmark (all cells)", CmdUnMark, F_NOARG|F_NORUN|F_UPDATE},
	{"update (view to be current)", CmdUpdate, F_NOARG},
	{"variables (are listed)", CmdVariables, F_NOARG|F_UPDATE|F_ABBR},
	{"wait (for computations)", CmdWait, F_NOARG},
	{"write (current object to) file", CmdWrite, F_ARG1|F_NORUN|F_ABBR},
	{"zero (current object)", CmdZero, F_NOARG|F_NORUN|F_UPDATE|F_ABBR},
	{0, 0, 0}				/* ends the table */
};


/*
 * The filename prefix.
 */
static	char	prefix[MAXPREFIX + 1];


static	int	FindCommand PROTO((char *, CMDTAB **));


void
ExpandFileName(buf, len)
	char *	buf;
	int	len;
{
	char *	entry;
	int	i;
	int	prefixlen;
	LIST	names;
	char	newpath[STRINGSIZE];

	InitList(&names);

	if (buf[0] == '/')
	{
		/*
		 * The buffer contains an absolute path,
		 * so only look there.
		 */
		GetCompletionFiles(&names, "/", buf);
	}
	else if (prefix[0] == '/')
	{
		/*
		 * There was an absolute prefix set, so only
		 * collect names from there.
		 */
		GetCompletionFiles(&names, prefix, buf);
	}
	else if (prefix[0] != '\0')
	{
		/*
		 * There was a relative prefix set.
		 * Append that path to all the library directory paths.
		 */
		GetCompletionFiles(&names, prefix, buf);

		for (i = 0; i < userlibcount; i++)
		{
			strcpy(newpath, userlibs[i]);
			strcat(newpath, "/");
			strcat(newpath, prefix);

			GetCompletionFiles(&names, newpath, buf);
		}

		strcpy(newpath, LIFELIB);
		strcat(newpath, "/");
		strcat(newpath, prefix);

		GetCompletionFiles(&names, newpath, buf);
	}
	else
	{
		/*
		 * There is no prefix at all.
		 * Look in the library paths for files.
		 */
		GetCompletionFiles(&names, ".", buf);

		for (i = 0; i < userlibcount; i++)
			GetCompletionFiles(&names, userlibs[i], buf);

		GetCompletionFiles(&names, LIFELIB, buf);
	}

	/*
	 * If there are no files then beep and let the user fix it.
	 */
	if (names.argc == 0)
	{
		Beep();

		return;
	}

	SortList(&names);

	/*
	 * Get a pointer to the last file entry in the buffer.
	 * This is the one that we will modify based on the list results.
	 */
	entry = strrchr(buf, '/');

	if (entry)
		entry++;
	else
		entry = buf;

	/*
	 * If there is exactly one file then copy the entry into the buffer
	 * possibly along with an extra space to indicate that the name
	 * is complete.
	 */
	if (names.argc == 1)
	{
		strcpy(entry, names.argv[0]);

		if (entry[0] && (entry[strlen(entry) - 1] != '/'))
			strcat(entry, " ");

		FreeList(&names);

		return;
	}

	/*
	 * There are multiple entries in the returned list.
	 * Get the length of the common prefix for all of those entries.
	 */
	prefixlen = GetListPrefixLength(&names);

	/*
	 * If there is no prefix in common, then just display the list
	 * of entries so the user can see the one he wants.
	 */
	if (prefixlen == 0)
	{
		DisplayList(&names, FALSE);

		FreeList(&names);

		return;
	}

	/*
	 * There is a prefix, so put that many characters into the
	 * buffer so that the user can see a partial match.  Do not
	 * append a space after the entry so he knows there are multiple
	 * matches.  Show the list of files anyway so it appears up to date.
	 */
	memcpy(entry, names.argv[0], prefixlen);
	entry[prefixlen] = '\0';

	DisplayList(&names, FALSE);

	FreeList(&names);
}


/*
 * Read and execute a line mode command.  This kind of command echoes,
 * and is terminated by an end of line.  Numeric arguments are available.
 * Some commands may have filename expansion done on them using a tab.
 */
void
DoLineCommand(arg1, arg2, got1, got2)
	VALUE	arg1;
	VALUE	arg2;
	BOOL	got1;
	BOOL	got2;
{
	CMDTAB *	cmd;		/* command structure */
	char *		name;		/* command name */
	char *		endname;
	char *		args;		/* arguments for command */
	char *		endargs;
	int		flag;		/* flags for command */
	int		matches;	/* number of matches found */
	char *		initbuf;	/* initial input buffer */
	int		len;
	int		endnamechar;
	BOOL		partial;

	initbuf = "";

	/*
	 * Iterate reading the command line until the line read is not a
	 * partial one (i.e, it is finally terminated by a newline.)
	 */
	for (;;)
	{
		/*
		 * Read a full or partial command line, and return a pointer
		 * to it.  Provide the data already read from the previous
		 * reading of a partial line, if there was one.
		 */
		initbuf = ReadPartialString("command: ",
			initbuf, strlen(initbuf));

		/*
		 * See if this line ends in a tab, and if so remember
		 * that this is a partial line and remove the tab.
		 */
		partial = FALSE;
		len = strlen(initbuf);

		if ((len > 0) && (initbuf[len - 1] == '\t'))
		{
			partial = TRUE;
			initbuf[--len] = '\0';
		}

		/*
		 * Skip leading blanks to find the beginning of the
		 * command name.
		 */
		name = initbuf;

		while (isblank(*name))
			name++;

		/*
		 * Find the end of the command name, save that terminating
		 * character, and temporarily replace it with a null so
		 * that we can look it up.
		 */
		endname = name;

		while (*endname && !isblank(*endname))
			endname++;

		endnamechar = *endname;
		*endname = '\0';

		/*
		 * Find the command name and remember which one this is.
		 * We don't yet cause an error for invalid commands
		 * since we may be in a partial command which could
		 * still be fixed up.
		 */
		matches = FindCommand(name, &cmd);

		/*
		 * Restore the character terminating the command,
		 * then skip blanks to find the beginning of the command
		 * arguments, if any.
		 */
		*endname = endnamechar;

		args = endname;

		while (isblank(*args))
			args++;

		/*
		 * If the line is not a partial line, then go execute it.
		 */
		if (!partial)
			break;

		/*
		 * This is a partial command line.
		 * If the command is valid and doesn't yet have a space
		 * after it, then complete the command name and add a
		 * following space to indicate that the command is ok.
		 */
		if (*name && (matches == 1) && (endnamechar == '\0'))
		{
			strcpy(name, cmd->name);

			endname = strchr(name, ' ');

			if (endname)
				*endname = '\0';

			strcat(name, " ");

			continue;
		}

		/*
		 * If the command is valid and the command allows filename
		 * recognition, then do that.
		 */
		if ((matches == 1) && (cmd->flags & F_EXPAND))
		{
			ExpandFileName(args, STRINGSIZE - (args - initbuf));

			continue;
		}

		/*
		 * The command is illegal or else it doesn't want recognition.
		 * Just beep and ignore the tab character.  The user can
		 * then edit the line to fix the problem.
		 */
		Beep();
	}

	/*
	 * The command line is complete.
	 * If it is an empty command, then just ignore it.
	 */
	if (*name == '\0')
		return;

	/*
	 * Complain about an invalid command if necessary.
	 */
	if (matches == 0)
		Error("Unknown line command");

	if (matches > 1)
		Error("Ambiguous line command");

	/*
	 * Strip trailing blanks from the last argument.
	 */
	endargs = initbuf + strlen(initbuf);

	while (isblank(endargs[-1]))
		endargs--;

	*endargs = '\0';

	/*
	 * Get the flags for the command and verify some things.
	 */
	flag = cmd->flags;

	if (flag & F_NORUN)
		CheckRun();

	if ((flag & F_ARG1) && (*args == '\0'))
		Error("Missing argument");

	if ((flag & F_NOARG) && *args)
		Error("Argument not allowed");

	/*
	 * Call the function with the arguments.
	 */
	(*cmd->func)(args, arg1, arg2, got1, got2);

	/*
	 * Fix up the view as specified by the command.
	 */
	if (flag & F_STATUS)
		curobj->update |= U_STAT;

	if (flag & F_UPDATE)
		curobj->update |= U_ALL;

	if (flag & F_REDRAW)
		curobj->update |= U_REDRAW;
}


/*
 * Search the specified string for a command match, and indirectly return
 * a pointer to the found command, if any.  Returns the number of commands
 * which match.  Zero indicates no match was found, and more than one
 * indicates the match was ambiguous.
 */ 
static int
FindCommand(name, retcmd)
	char *		name;
	CMDTAB **	retcmd;
{
	CMDTAB *	cmd;		/* command structure */
	CMDTAB *	cmdworks;	/* command that was found */
	CMDTAB *	cmdbest;	/* command found that is wanted */
	int		matches;	/* number of matches found */

	*retcmd = 0;
	matches = 0;

	/*
	 * Seatch the table for the command.
	 * Exact match always wins, followed by an indicated abbreviation,
	 * followed by a unique abbreviation.
	 */
	cmdbest = NULL;
	cmdworks = NULL;

	for (cmd = cmdtab; cmd->name != NULL; cmd++)
	{
		if (strcmp(name, cmd->name) == 0)
		{
			cmdbest = cmd;

			break;
		}

		if (!IsAbbreviation(name, cmd->name))
			continue;

		cmdworks = cmd;
		matches++;

		if (cmd->flags & F_ABBR)
			cmdbest = cmd;
	}

	/*
	 * If we found an exact match or an abbreviation of an
	 * important command, then return that one unambiguously.
	 */
	if (cmdbest)
	{
		*retcmd = cmdbest;

		return 1;
	}

	/*
	 * Return the found command if any, and the number of matches.
	 */
	*retcmd = cmdworks;

	return matches;
}


/*
 * Advance marked cells by the specified number of generations.
 */
static void
CmdAdvance(cp, arg1, arg2, got1, got2)
	char *	cp;
	VALUE	arg1;
	VALUE	arg2;
	BOOL	got1;
	BOOL	got2;
{
	COUNT	count;

	count = 1;

	if (*cp)
	{
		count = GetExpression(cp);

		if (count <= 0)
			return;
	}

	Backup();

	CopyMarkedObject(curobj, tempobject, curobj->curmarks);
	MoveMarkedObject(curobj, deleteobject, curobj->curmarks);

	genleft = count;

	while ((genleft > 0) && !stop)
		DoGeneration(tempobject);

	curobj->cellmarks = curobj->curmarks;
	AddObject(tempobject, curobj, RELATIVE, FALSE);
	curobj->cellmarks = MARK_ANY;
}


/*
 * Assign a button to a macro.
 */
static void
CmdAssign(cp, arg1, arg2, got1, got2)
	char *	cp;
	VALUE	arg1;
	VALUE	arg2;
	BOOL	got1;
	BOOL	got2;
{
	VALUE	button;
	int	macro;

	button = 0;

	while (isdigit(*cp))
		button = button * 10 + *cp++ - '0';

	if (*cp && !isblank(*cp))
		Error("Illegal button value");

	while (isblank(*cp))
		cp++;

	macro = *cp;

	if (macro)
		cp++;

	while (isblank(*cp))
		cp++;

	if (*cp || ((macro != '\0') && !MacroIsValid(macro)))
		Error("Illegal macro name");

	if (!(*dev->assign)(dev, button, macro))
		Error("Button is not assignable");
}


/*
 * Clear the path from the current object.
 */
static void
CmdClearPath(cp, arg1, arg2, got1, got2)
	char *	cp;
	VALUE	arg1;
	VALUE	arg2;
	BOOL	got1;
	BOOL	got2;
{
	Backup();
	ClearPath(curobj);
}


/*
 * Shorten the path from the current object by one or more segments.
 */
static void
CmdShortenPath(cp, arg1, arg2, got1, got2)
	char *	cp;
	VALUE	arg1;
	VALUE	arg2;
	BOOL	got1;
	BOOL	got2;
{
	COUNT	count;

	count = 1;

	if (*cp)
		count = GetExpression(cp);

	if (count <= 0)
		Error("Bad shorten count value");

	Backup();

	ShortenPath(curobj, count);
}


/*
 * Restore the path that was set for an object.
 */
static void
CmdRestorePath(cp, arg1, arg2, got1, got2)
	char *	cp;
	VALUE	arg1;
	VALUE	arg2;
	BOOL	got1;
	BOOL	got2;
{
	Backup();

	if (!RestorePath(curobj))
		Error("There is no path to restore");
}


/*
 * Specify the group of cells acted on by marks.
 */
static void
CmdGroups(cp, arg1, arg2, got1, got2)
	char *	cp;
	VALUE	arg1;
	VALUE	arg2;
	BOOL	got1;
	BOOL	got2;
{
	MARK	mark;

	mark = 0;

	while (*cp)
	{
		if (!isletter(*cp))
			Error("Illegal group character");

		mark |= MARK_GROUPS << (*cp++ - 'a');
	}

	if (mark != curobj->curmarks)
	{
		curobj->curmarks = mark;
		curobj->update |= U_ALL;
	}
}


/*
 * Copy the current object into another object.
 */
static void
CmdCopy(cp, arg1, arg2, got1, got2)
	char *	cp;
	VALUE	arg1;
	VALUE	arg2;
	BOOL	got1;
	BOOL	got2;
{
	CopyObject(curobj, GetObject(cp));
}


/*
 * Copy the currently marked cells into another object.
 */
static void
CmdCopyMarked(cp, arg1, arg2, got1, got2)
	char *	cp;
	VALUE	arg1;
	VALUE	arg2;
	BOOL	got1;
	BOOL	got2;
{
	CopyMarkedObject(curobj, GetObject(cp), curobj->curmarks);
}


/*
 * Get the current object from another object.
 * This totally destroys the current object.
 */
static void
CmdGet(cp, arg1, arg2, got1, got2)
	char *	cp;
	VALUE	arg1;
	VALUE	arg2;
	BOOL	got1;
	BOOL	got2;
{
	OBJECT *obj;

	obj = FindObject(cp);

	if (obj == NULL)
		Error("No such object");

	CopyObject(obj, curobj);
}


/*
 * Set a default value for new objects.
 */
static void
CmdDefault(cp, arg1, arg2, got1, got2)
	char *	cp;
	VALUE	arg1;
	VALUE	arg2;
	BOOL	got1;
	BOOL	got2;
{
	char *	param;
	VALUE	val;

	param = cp;

	while (*cp && !isblank(*cp))
		cp++;

	while (isblank(*cp))
		*cp++ = '\0';

	val = 1;

	if (IsAbbreviation(param, "scale"))
	{
		if (*cp)
			val = GetExpression(cp);

		if ((val == 0) || (val > MAXSCALE))
			Error("Bad scale value");

		if (val == -1)
			val = 1;

		defaultscale = val;

		return;
	}

	if (IsAbbreviation(param, "frequency"))
	{
		if (*cp)
			val = GetExpression(cp);

		if (val <= 0)
			Error("Illegal frequency value");

		defaultfrequency = val;

		return;
	}

	if (IsAbbreviation(param, "tracks"))
	{
		if (IsAbbreviation(cp, "on"))
			defaulttracks = TRUE;
		else if (IsAbbreviation(cp, "off"))
			defaulttracks = FALSE;
		else
			Error("Bad tracks default");

		return;
	}

	Error("Illegal parameter for defaulting");
}


/*
 * Manipulate the tracking of old cells.
 */
static void
CmdTracks(cp, arg1, arg2, got1, got2)
	char *	cp;
	VALUE	arg1;
	VALUE	arg2;
	BOOL	got1;
	BOOL	got2;
{
	if ((*cp == '\0') || IsAbbreviation(cp, "on"))
	{
		curobj->tracks = TRUE;

		return;
	}

	if (IsAbbreviation(cp, "off"))
	{
		curobj->tracks = FALSE;

		return;
	}

	if (IsAbbreviation(cp, "clear"))
	{
		if (curobj->usedrows == termrow)
			return;

		FreeRows(curobj->usedrows);
		curobj->usedrows = termrow;
		curobj->update |= U_ALL;

		return;
	}

	if (IsAbbreviation(cp, "set"))
	{
		SaveUsedCells(curobj);
		curobj->update |= U_ALL;

		return;
	}

	Error("Bad tracks argument");
}


/*
 * Destroy an existing object.
 */
static void
CmdDestroy(cp, arg1, arg2, got1, got2)
	char *	cp;
	VALUE	arg1;
	VALUE	arg2;
	BOOL	got1;
	BOOL	got2;
{
	OBJECT *obj;

	obj = FindObject(cp);

	if (obj == NULL)
		Error("No such object");

	DestroyObject(obj);
}


/*
 * Dump things that the user might want initialized at startup to a file.
 * These are things like the current macros and button assignments.
 */
static void
CmdDumpInitData(cp, arg1, arg2, got1, got2)
	char *	cp;
	VALUE	arg1;
	VALUE	arg2;
	BOOL	got1;
	BOOL	got2;
{
	FILE *	fp;

	fp = fopen(cp, "w");

	if (fp == NULL)
		Error("Cannot create file");

	WriteMacros(fp);
	(*dev->writeassign)(dev, fp);

	fflush(fp);

	if (ferror(fp))
	{
		fclose(fp);
		Error("Write failed");
	}

	if (fclose(fp))
		Error("Close failed");
}


/*
 * Edit an object.  A null argument implies the previous object.
 */
static void
CmdEdit(cp, arg1, arg2, got1, got2)
	char *	cp;
	VALUE	arg1;
	VALUE	arg2;
	BOOL	got1;
	BOOL	got2;
{
	OBJECT *obj;

	obj = prevobj;

	if (*cp)
	{
		obj = FindObject(cp);

		if (obj == NULL)
			obj = GetObject(cp);
	}

	if (obj == NULL)
		Error("No such object");

	SetObject(obj);
}


/*
 * Undo the last change made to the current object.
 */
static void
CmdUndo(cp, arg1, arg2, got1, got2)
	char *	cp;
	VALUE	arg1;
	VALUE	arg2;
	BOOL	got1;
	BOOL	got2;
{
	MoveObject(curobj, tempobject);
	MoveObject(backupobject, curobj);
	MoveObject(tempobject, backupobject);
}


/*
 * End current input level.
 */
static void
CmdEndInputLevel(cp, arg1, arg2, got1, got2)
	char *	cp;
	VALUE	arg1;
	VALUE	arg2;
	BOOL	got1;
	BOOL	got2;
{
	if (curinput <= inputs)
		Error("Cannot end top level input");

	(*curinput->term)(curinput);
}


/*
 * Update the view even if inside of a loop or macro.
 */
static void
CmdUpdate(cp, arg1, arg2, got1, got2)
	char *	cp;
	VALUE	arg1;
	VALUE	arg2;
	BOOL	got1;
	BOOL	got2;
{
	curobj->update |= U_ALL;	/* IS THIS NECESSARY?? */
	UpdateView();
}


/*
 * Wait until all outstanding generation computations are finished.
 */
static void
CmdWait(cp, arg1, arg2, got1, got2)
	char *	cp;
	VALUE	arg1;
	VALUE	arg2;
	BOOL	got1;
	BOOL	got2;
{
	while (!stop && (genleft > 0))
	{
		DoGeneration(curobj);
		UpdateView();
	}
}


/*
 * Set output frequency for object.
 */
static void
CmdFrequency(cp, arg1, arg2, got1, got2)
	char *	cp;
	VALUE	arg1;
	VALUE	arg2;
	BOOL	got1;
	BOOL	got2;
{
	COUNT	freq;

	freq = 1;

	if (*cp)
		freq = GetExpression(cp);

	if (freq <= 0)
		Error("Illegal frequency value");

	curobj->frequency = freq;
	freqcount = freq;
}


/*
 * Set the mode associated with movement.
 */
static void
CmdMode(cp, arg1, arg2, got1, got2)
	char *	cp;
	VALUE	arg1;
	VALUE	arg2;
	BOOL	got1;
	BOOL	got2;
{
	if ((*cp == '\0') || IsAbbreviation(cp, "move"))
	{
		mode = M_MOVE;

		return;
	}

	if (IsAbbreviation(cp, "insert"))
	{
		mode = M_INSERT;

		return;
	}

	if (IsAbbreviation(cp, "delete"))
	{
		mode = M_DELETE;

		return;
	}

	Error("Illegal movement mode");
}


/*
 * Unmark all marks which are visible.
 */
static void
CmdUnMark(cp, arg1, arg2, got1, got2)
	char *	cp;
	VALUE	arg1;
	VALUE	arg2;
	BOOL	got1;
	BOOL	got2;
{
	ClearMarks(curobj, curobj->curmarks);
}


/*
 * Invert the marks for all cells.
 */
static void
CmdInvertMarks(cp, arg1, arg2, got1, got2)
	char *	cp;
	VALUE	arg1;
	VALUE	arg2;
	BOOL	got1;
	BOOL	got2;
{
	InvertMarks(curobj, curobj->curmarks);
}


/*
 * Select the character used for the background of the screen.
 */
static void
CmdGrid(cp, arg1, arg2, got1, got2)
	char *	cp;
	VALUE	arg1;
	VALUE	arg2;
	BOOL	got1;
	BOOL	got2;
{
	if (*cp == '\0')
		cp = ".";

	if ((*cp < ' ') || (cp[1] != '\0'))
		Error("Bad grid character");

	gridchar = *cp;
}


/*
 * Set so we don`t see a grid on the screen.
 */
static void
CmdNoGrid(cp, arg1, arg2, got1, got2)
	char *	cp;
	VALUE	arg1;
	VALUE	arg2;
	BOOL	got1;
	BOOL	got2;
{
	gridchar = ' ';
}


/*
 * Specify what the shape used for drawing cells is.
 * This is either squares or circles.
 */
static void
CmdShape(cp, arg1, arg2, got1, got2)
	char *	cp;
	VALUE	arg1;
	VALUE	arg2;
	BOOL	got1;
	BOOL	got2;
{
	if (*cp == '\0')
		drawcircles = !drawcircles;
	else if (IsAbbreviation(cp, "circles"))
		drawcircles = TRUE;
	else if (IsAbbreviation(cp, "squares"))
		drawcircles = FALSE;
	else
		Error("Shape command needs 'circles' or 'squares'");
}


/*
 * Display a list of the line mode commands.
 */
static void
CmdHelp(cp, arg1, arg2, got1, got2)
	char *	cp;
	VALUE	arg1;
	VALUE	arg2;
	BOOL	got1;
	BOOL	got2;
{
	CMDTAB *	cmd;
	int		count;
	char		buf[80];

	ShowHelp("\
The following table lists all line mode commands.  Unique abbreviations are\n\
allowed.  Commands shown with '*' can be abbreviated even when ambiguous.\n");

	count = 0;

	for (cmd = cmdtab; cmd->name; cmd++)
	{
		if ((count++ % 2) == 0)
			ShowHelp("\n");

		sprintf(buf, "%c %-35s", ((cmd->flags & F_ABBR) ? '*' : ' '),
			cmd->name);

		ShowHelp(buf);
	}

	ShowHelp("\n");
	EndHelp(TRUE);
}


/*
 * Display a list of the *.l files that can be read.
 * If a pattern is given, only list files matching that pattern.
 * If a prefix is set, then that is taken into account.
 */
static void
CmdList(cp, arg1, arg2, got1, got2)
	char *	cp;
	VALUE	arg1;
	VALUE	arg2;
	BOOL	got1;
	BOOL	got2;
{
	int	i;
	char *	dirname;
	LIST	names;
	char	buf[MAXPATH];

	InitList(&names);

	/*
	 * Look for an option, which must be -d.
	 * If found, then just look in that particular directory.
	 */
	if (*cp == '-')
	{
		dirname = cp;

		while (*cp && !isblank(*cp))
			cp++;

		if ((cp != dirname + 2) || (memcmp(dirname, "-d", 2) != 0))
			Error("Bad option for list command");

		while (isblank(*cp))
			cp++;

		dirname = cp;

		if (*dirname == '\0')
			dirname = ".";

		while (*cp && !isblank(*cp))
			cp++;

		if (*cp)
			*cp++ = '\0';

		while (isblank(*cp))
			cp++;

		GetFiles(&names, dirname, cp, strlen(dirname) + 1, 1);
	}
	else if (prefix[0] == '/')
	{
		/*
		 * There was an absolute prefix set, so only look there.
		 */
		GetFiles(&names, prefix, cp, strlen(prefix) + 1, MAXLIST);
	}
	else if (prefix[0] != '\0')
	{
		/*
		 * There was a relative prefix set.
		 * Append that path to all the library directory paths.
		 */
		GetFiles(&names, prefix, cp, strlen(prefix) + 1, 1);

		for (i = 0; i < userlibcount; i++)
		{
			strcpy(buf, userlibs[i]);
			strcat(buf, "/");
			strcat(buf, prefix);

			GetFiles(&names, buf, cp, strlen(buf) + 1, MAXLIST);
		}

		strcpy(buf, LIFELIB);
		strcat(buf, "/");
		strcat(buf, prefix);

		GetFiles(&names, buf, cp, strlen(buf) + 1, MAXLIST);
	}
	else
	{
		/*
		 * There is no prefix at all.
		 * Look in the library paths for files.
		 */
		GetFiles(&names, ".", cp, 2, 1);

		for (i = 0; i < userlibcount; i++)
		{
			GetFiles(&names, userlibs[i], cp,
				strlen(userlibs[i]) + 1, MAXLIST);
		}

		GetFiles(&names, LIFELIB, cp, strlen(LIFELIB) + 1, MAXLIST);
	}

	if (names.argc == 0)
		Error("No files found");

	DisplayList(&names, TRUE);

	FreeList(&names);
}


/*
 * Insert another object into this one.
 * This preserves existing cells in the current object.
 */
static void
CmdInsert(cp, arg1, arg2, got1, got2)
	char *	cp;
	VALUE	arg1;
	VALUE	arg2;
	BOOL	got1;
	BOOL	got2;
{
	OBJECT *obj;

	obj = FindObject(cp);

	if (obj == NULL)
		Error("No such object");

	Backup();

	curobj->cellmarks = curobj->curmarks;
	AddObject(obj, curobj, RELATIVE, FALSE);
	curobj->cellmarks = MARK_ANY;
}


/*
 * Show a list of objects.
 */
static void
CmdObjects(cp, arg1, arg2, got1, got2)
	char *	cp;
	VALUE	arg1;
	VALUE	arg2;
	BOOL	got1;
	BOOL	got2;
{
	BOOL	all;

	all = ((*cp == '-') || (*cp == 'a'));

	ListObjects(all);
}


/*
 * Rename the current object to something else.
 */
static void
CmdRename(cp, arg1, arg2, got1, got2)
	char *	cp;
	VALUE	arg1;
	VALUE	arg2;
	BOOL	got1;
	BOOL	got2;
{
	if (curobj->reserved)
		Error("Cannot rename reserved object");

	if (strlen(cp) > MAXNAME)
		Error("Name too long");

	if (FindObject(cp))
		Error("Name already exists");

	if (BADNAME(cp))
		Error("Cannot create reserved name");

	strcpy(curobj->name, cp);
}


/*
 * Move current object into another object.
 */
static void
CmdMove(cp, arg1, arg2, got1, got2)
	char *	cp;
	VALUE	arg1;
	VALUE	arg2;
	BOOL	got1;
	BOOL	got2;
{
	MoveObject(curobj, GetObject(cp));
}


/*
 * Move currently markeded cells into another object.
 */
static void
CmdMoveMarked(cp, arg1, arg2, got1, got2)
	char *	cp;
	VALUE	arg1;
	VALUE	arg2;
	BOOL	got1;
	BOOL	got2;
{
	MoveMarkedObject(curobj, GetObject(cp), curobj->curmarks);
}


/*
 * Set the value of a variable.
 */
static void
CmdSet(cp, arg1, arg2, got1, got2)
	char *	cp;
	VALUE	arg1;
	VALUE	arg2;
	BOOL	got1;
	BOOL	got2;
{
	char *	exp;

	for (exp = cp; *exp && !isblank(*exp) && (*exp != '='); exp++)
		;

	/*
	 * Skip spaces.
	 */
	if (isblank(*exp))
	{
		*exp++ = '\0';

		while (isblank(*exp))
			exp++;
	}

	/*
	 * If there is no expression, then set the variable to zero.
	 */
	if (*exp == '\0')
	{
		SetVariable(cp, 0);

		return;
	}

	if (*exp == '=')
		*exp++ = '\0';

	SetVariable(cp, GetExpression(exp));
}


/*
 * Type the value of an expression.
 */
static void
CmdType(cp, arg1, arg2, got1, got2)
	char *	cp;
	VALUE	arg1;
	VALUE	arg2;
	BOOL	got1;
	BOOL	got2;
{
	static	char	buf[30];

	sprintf(buf, "Value = %ld", GetExpression(cp));
	errorstring = buf;
}


/*
 * Set the delay time between updates in milliseconds.
 */
static void
CmdDelay(cp, arg1, arg2, got1, got2)
	char *	cp;
	VALUE	arg1;
	VALUE	arg2;
	BOOL	got1;
	BOOL	got2;
{
	long	millisec;

	if (*cp == '\0')
	{
		delaytime = 0;

		return;
	}

	millisec = GetExpression(cp);

	if ((millisec < 0) || (millisec > MAXDELAY))
		Error("Illegal delay time");

	delaytime = millisec;
}


/*
 * Display the values of all the variables.
 */
static void
CmdVariables(cp, arg1, arg2, got1, got2)
	char *	cp;
	VALUE	arg1;
	VALUE	arg2;
	BOOL	got1;
	BOOL	got2;
{
	ListVariables();
}


/*
 * Quit program
 */
static void
CmdQuit(cp, arg1, arg2, got1, got2)
	char *	cp;
	VALUE	arg1;
	VALUE	arg2;
	BOOL	got1;
	BOOL	got2;
{
	(*dev->close)(dev);

	exit(0);
}


/*
 * Set or display the path prefix for reading and listing files.
 */
static void
CmdPrefix(cp, arg1, arg2, got1, got2)
	char *	cp;
	VALUE	arg1;
	VALUE	arg2;
	BOOL	got1;
	BOOL	got2;
{
	int	len;
	static	char	buf[MAXPREFIX + 20];

	/*
	 * If there is no argument then display the current prefix.
	 */
	if (*cp == '\0')
	{
		if (prefix[0])
		{
			sprintf(buf, "Prefix = %s", prefix);
			errorstring = buf;
		}
		else
			errorstring = "No prefix";

		curobj->update |= U_STAT;

		return;
	}

	/*
	 * If the prefix is just a dot then clear it.
	 */
	if ((cp[0] == '.') && (cp[1] == '\0'))
	{
		prefix[0] = '\0';

		return;
	}

	/*
	 * If the prefix is too long then complain.
	 */
	len = strlen(cp);

	if (len > MAXPREFIX)
		Error("Prefix string is too long");

	/*
	 * Remove trailing slashes except for an initial one,
	 * and then save the prefix away for future use.
	 */
	while ((len > 1) && (cp[len - 1] == '/'))
		cp[--len] = '\0';

	strcpy(prefix, cp);
}


/*
 * Read commands or object from file, defaulting extension if needed.
 * If a prefix is set, then apply that prefix to the name.
 */
static void
CmdRead(cp, arg1, arg2, got1, got2)
	char *	cp;
	VALUE	arg1;
	VALUE	arg2;
	BOOL	got1;
	BOOL	got2;
{
	char	buf[MAXPATH];

	Backup();

	/*
	 * If there is a prefix then apply it.
	 */
	if (prefix[0])
	{
		if (strlen(prefix) + strlen(cp) + 2 > MAXPATH)
			Error("Path is too long");

		strcpy(buf, prefix);

		if ((prefix[0] != '/') || (prefix[1] != '\0'))
			strcat(buf, "/");

		strcat(buf, cp);

		cp = buf;
	}
	
	if (!SetFile(cp))
		Error("Cannot open input file");
}


/*
 * Set new life rule.
 */
static void
CmdRule(cp, arg1, arg2, got1, got2)
	char *	cp;
	VALUE	arg1;
	VALUE	arg2;
	BOOL	got1;
	BOOL	got2;
{
	if (*cp == '\0')
		cp = DEFAULTRULE;

	SetRule(cp);
}


/*
 * Read commands from the terminal.  Useful in loops or command files.
 * If the -c argument is given, we don't do it if no input is ready.
 */
static void
CmdTtyInput(cp, arg1, arg2, got1, got2)
	char *	cp;
	VALUE	arg1;
	VALUE	arg2;
	BOOL	got1;
	BOOL	got2;
{
	/*
	 * Possibly check for input.
	 */
	if ((*cp == '-') || (*cp == 'c'))
	{
		if (!(*dev->inputready)(dev))
			return;
	}

	if (!SetTty())
		Error("Nesting too deep");
}


/*
 * Write object to file.
 */
static void
CmdWrite(cp, arg1, arg2, got1, got2)
	char *	cp;
	VALUE	arg1;
	VALUE	arg2;
	BOOL	got1;
	BOOL	got2;
{
	char *	method;

	if (!got1)
		arg1 = WRITECOLS;

	method = "rle";

	if (*cp == '-')
	{
		method = cp + 1;

		while (*cp && !isblank(*cp))
			cp++;

		while (isblank(*cp))
			*cp++ = '\0';
	}

	WriteObject(curobj, method, cp, arg1 ? WRITEROWS : 0, arg1);
}


/*
 * Zero out current object.
 */
static void
CmdZero(cp, arg1, arg2, got1, got2)
	char *	cp;
	VALUE	arg1;
	VALUE	arg2;
	BOOL	got1;
	BOOL	got2;
{
	OBJECT *obj;

	obj = curobj;

	Backup();

	ZeroObject(obj);

	obj->gen = 0;
	obj->born = 0;
	obj->died = 0;
	obj->currow = 0;
	obj->curcol = 0;
	obj->origrow = 0;
	obj->origcol = 0;
	obj->autoscale = FALSE;

	memset(obj->wherelocs, 0, sizeof(obj->wherelocs));

	SetScale(obj, defaultscale);

	mode = M_MOVE;
	obj->frequency = defaultfrequency;
	obj->tracks = defaulttracks;
	freqcount = defaultfrequency;
}


/*
 * See if one string is an abbreviation of another.
 * Returns TRUE if first string is an abbreviation.
 */
BOOL
IsAbbreviation(str1, str2)
	char *	str1;
	char *	str2;
{
	if ((str1 == NULL) || (str2 == NULL))
		return FALSE;

	while (*str1)
	{
		if (*str1++ != *str2++)
			return FALSE;
	}

	return TRUE;
}

/* END CODE */
